#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h> 
#include <stdint.h>
#include "rets.h"

typedef long (*rec_fun_t)(uint8_t flags, uint8_t* buffer, long offset, long size);

//Ugly but, meh
char* output_file = NULL;

void* read_file(char* filename, long* size)
{
    FILE* f = fopen(filename, "rb");
    void* buffer = NULL;
    long loc_size = 0;

    if ( ! f)
    {
        fprintf(stderr, "Cannot open input file %s: %s (%d)\n", filename, strerror(errno), errno);
        goto error;
    }

    if (fseek(f, 0, SEEK_END))
    {
        fprintf(stderr, "Cannot seek to the end of the input file %s: %s (%d)\n", filename, strerror(errno), errno);
        goto error;
    }

    if ( (loc_size = ftell(f)) == -1)
    {
        fprintf(stderr, "Cannot tell the size of the input file %s: %s (%d)\n", filename, strerror(errno), errno);
        goto error;
    } 

    if (fseek(f, 0, SEEK_SET))
    {
        fprintf(stderr, "Cannot seek back to the beginning of the input file %s: %s (%d)\n", filename, strerror(errno), errno);
        goto error;
    } 

    if ( ! (buffer = malloc(loc_size)) )
    {
        fprintf(stderr, "Cannot allocate %lu bytes of memory: %s (%d)\n", loc_size, strerror(errno), errno);
        goto error;        
    }

    if (fread(buffer, loc_size, 1, f) != 1)
    {
        fprintf(stderr, "Cannot read from the input file %s: %s (%d)\n", filename, strerror(errno), errno);
        goto error;        
    }    

    fclose(f);

    *size = loc_size;
    return buffer;

error:
    if (f)
        fclose(f);
    if (buffer)
        free(buffer);
    return NULL;    
}

#define NEXT(type, value, buffer, offset, size) \
{ \
    if ((unsigned long)offset > size -  sizeof(type)) \
    { \
        fprintf(stderr, "Input file is too small, tried to read %ld bytes at offset %lx but the size is only %ld.\n", sizeof(type), offset, size); \
        exit(2); \
    } \
    type value = *(type*)(buffer + offset);  \
    (offset) += sizeof(type);  \
    return value; \
}


uint8_t next_byte(uint8_t* buffer, long* offset, long size) 
{
    NEXT(uint8_t, value, buffer, *offset, size);
}

uint16_t next_word(uint8_t* buffer, long* offset, long size) 
{
    NEXT(uint16_t, value, buffer, *offset, size);
}

uint32_t next_dword(uint8_t* buffer, long* offset, long size) 
{
    NEXT(uint32_t, value, buffer, *offset, size);
}

//Helpers which assume specific variable names
#define NEXT_BYTE() next_byte(buffer, &offset, size)
#define NEXT_WORD() next_word(buffer, &offset, size)
#define NEXT_DWORD() next_dword(buffer, &offset, size)

//Parser
#define xfLMOVE     0x08
#define xfAUTO      0x01
#define xfEMBELL    0x02
#define xfNULL      0x01
#define xfLSPACE    0x04
#define xfRULER     0x02

long parse_mtfe(uint8_t* buffer, long offset, long size, rec_fun_t* records);

long parse_nudge(uint8_t* buffer, long offset, long size, int16_t* dx, int16_t* dy)
{
    *dx = NEXT_BYTE();
    *dy = NEXT_BYTE();

    if (*dx == 0x80 && *dy == 0x80)
    {
        *dx = (int16_t)NEXT_WORD();
        *dy = (int16_t)NEXT_WORD();
    }

    return offset;
}

long handle_nudge(uint8_t flags, uint8_t* buffer, long offset, long size, int16_t* dx, int16_t* dy)
{
    int16_t loc_dx, loc_dy;
    int16_t* ptr_dx = dx ? dx : &loc_dx, *ptr_dy = dy ? dy : &loc_dy;

    if (flags & xfLMOVE)
    {
        offset = parse_nudge(buffer, offset, size, ptr_dx, ptr_dy);
        printf("\tNudge values:\n\t\tdx: %hd\n\t\tdy: %hd\n", *ptr_dx, *ptr_dy);
    }
    return offset;
}

long handle_linespacing(uint8_t flags, uint8_t* buffer, long offset, long size, uint8_t* ls)
{
    if (flags & xfLSPACE)
    {
        uint8_t loc_ls;
        uint8_t* ptr = ls ? ls : &loc_ls;

        *ptr = NEXT_BYTE();
        printf("\tLine spacing: %d\n", *ptr);
    }
    return offset;
}

struct tabstop_t {
    uint8_t type;
    uint16_t offset;
};

struct tabstops_t {
    struct tabstop_t tabs[256];
};

long handle_ruler(uint8_t flags, uint8_t* buffer, long offset, long size, uint8_t* ntabs, struct tabstops_t* tabs)
{
    char* stops[] = {"left", "center", "right", "equal", "decimal"};

    if (flags & xfRULER)
    {
        printf("\tTab stops:\n");

        uint8_t loc_ntabs;
        struct tabstops_t loc_tabs;

        uint8_t* ptr_ntabs = ntabs ? ntabs : &loc_ntabs;
        struct tabstops_t* ptr_tabs = tabs ? tabs : &loc_tabs;

        *ptr_ntabs = NEXT_BYTE();

        for (uint8_t i = 0; i < *ptr_ntabs; i++)
        {
            struct tabstop_t tab;
            tab.type = NEXT_BYTE();
            tab.offset = NEXT_WORD();

            printf("\t\ttype: %s\n\t\tvalue: %hd\n", stops[tab.type], tab.offset);
            ptr_tabs->tabs[i] = tab;
        }

    }
    return offset;
}

long parse_mtfe(uint8_t* buffer, long offset, long size, rec_fun_t* records);
long end_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long line_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long char_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long template_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long pile_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long matrix_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long embel_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long ruler_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long font_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long size_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long full_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long sub_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long sub2_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long sym_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long subsym_record(uint8_t flags, uint8_t* buffer, long offset, long size);
long unk_record(uint8_t flags, uint8_t* buffer, long offset, long size);

rec_fun_t stdrec[] = {
    end_record,     line_record,    char_record,    template_record, 
    pile_record,    matrix_record,  embel_record,   ruler_record,
    font_record,    size_record,    full_record,    sub_record,
    sub2_record,    sym_record,     subsym_record,  unk_record
};


char* haligns[] = {"UNK", "left justification", "centered", "right justification", 
    "relational operator alignment", "decimal point alignment"};
char* valigns[] = {"top line", "center line", "bottom line", "vertical centering"};

long end_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags; (void)buffer; (void)offset; (void)size;
    printf("END record\n");
    return -1;
}
long line_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    printf("LINE record\n");
    offset = handle_nudge(flags, buffer, offset, size, NULL, NULL);
    offset = handle_linespacing(flags, buffer, offset, size, NULL);
    offset = handle_ruler(flags, buffer, offset, size, NULL, NULL);
    return parse_mtfe(buffer, offset, size, stdrec);
}

long char_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    char* typefaces[] = {
        "UNK", "fnTEXT", "fnFUNCTION","fnVARIABLE","fnLCGREEK","fnUCGREEK",
        "fnSYMBOL","fnVECTOR","fnNUMBER","fnUSER1","fnUSER2","fnMTEXTRA",
         "fnTEXT_FE"
    };

    printf("CHAR record\n");
    offset = handle_nudge(flags, buffer, offset, size, NULL, NULL);
    uint8_t typeface = NEXT_BYTE();

    if (typeface > 0x80 && typeface <= 0x8c)
        printf("\tType face: %s\n", typefaces[typeface]);
    else
        printf("\tType face: %d\n", typeface + 0x80);

    uint16_t char_point = NEXT_WORD();
    printf("\tCharacter value: 0x%04x (%d)\n", char_point, char_point);

    if (flags & xfEMBELL)
        offset = parse_mtfe(buffer, offset, size, stdrec);

    return offset;
}

long template_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    printf("TMPL record\n");
    offset = handle_nudge(flags, buffer, offset, size, NULL, NULL);
    uint8_t selector = NEXT_BYTE(), variation = NEXT_BYTE(),  options = NEXT_BYTE();
    printf("\tselector: %d\n\tvariation: %d\n\toptions: %d\n", selector, variation, options);
    return parse_mtfe(buffer, offset, size, stdrec);
}

long pile_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    

    printf("PILE record\n");
    offset = handle_nudge(flags, buffer, offset, size, NULL, NULL);
    uint8_t halign = NEXT_BYTE(), valign = NEXT_BYTE();
    printf("\thalign: %s (%d)\n\tvalign: %s (%d)\n", halign <= 5 ? haligns[halign] : "UNK", halign, valign <= 3 ? valigns[valign] : "UNK", valign);
    offset = handle_ruler(flags, buffer, offset, size, NULL, NULL);

    return parse_mtfe(buffer, offset, size, stdrec);
}

uint8_t print_separators(uint8_t sep, uint8_t left)
{
    char* separators[] = {"none", "solid", "dashed", "dotted"};
    uint8_t j = 6;
    while (j > 0 && left > 0)
    {
        printf("\t%s", separators[(sep >> j) & 0x3]);
        j -= 2;
        left--;
    }

    return left;
}

long matrix_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    char* haligns[] = {"UNK", "left justification", "centered", "right justification", 
    "relational operator alignment", "decimal point alignment"};
    char* valigns[] = {"top line", "center line", "bottom line", "vertical centering"};

    printf("MATRIX record\n");
    offset = handle_nudge(flags, buffer, offset, size, NULL, NULL);
    uint8_t valign = NEXT_BYTE(), h_just = NEXT_BYTE(), v_just = NEXT_BYTE();
    printf("\thalign: %s (%d)\n\th_just: %s (%d)\n\tv_just: %s (%d)\n", valign <= 3 ? valigns[valign] : "UNK", valign,
                h_just <= 5 ? haligns[h_just] : "UNK", h_just,  
                v_just <= 3 ? valigns[v_just] : "UNK", v_just);
    
    uint8_t rows = NEXT_BYTE(), cols = NEXT_BYTE();
    printf("\tdimensions: %dx%d (rows x cols)\n ", rows, cols);
    uint8_t n_rows_parts = (rows + 1) / 4 + ((rows + 1) % 4 ? 1 : 0), n_cols_parts = (cols + 1) / 4 + ((cols + 1) % 4 ? 1 : 0);

    uint8_t k = rows + 1;
    printf("\tRow partitions:\n\t");
    for (uint8_t i = 0; i < n_rows_parts; i++)
        k = print_separators(NEXT_BYTE(), k);
    printf("\n");

    k = cols + 1;
    printf("\tCols partitions:\n\t");
    for (uint8_t i = 0; i < n_cols_parts; i++)
        k = print_separators(NEXT_BYTE(), k);
    printf("\n");


    return parse_mtfe(buffer, offset, size, stdrec);
}

long embel_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    char* embells[] = {
        "UNK", "UNK", "embDOT single dot","embDDOT","embTDOT","embPRIME","embDPRIME",
        "embBPRIME","embTILDE","embHAT","embNOT","embRARROW","embLARROW",
        "embBARROW","embR1ARROW","embL1ARROW","embMBAR","embOBAR",
        "embTPRIME","embFROWN","embSMILE"
    };
    printf("EMBELL record\n");
    offset = handle_nudge(flags, buffer, offset, size, NULL, NULL);
    uint8_t embell = NEXT_BYTE();
    printf("\ttype: %s (%d)\n", embell <= 20 ? embells[embell] : "UNK", embell);
    return offset;
}

long ruler_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags;
    printf("RULER record\n");
    return handle_ruler(xfRULER, buffer, offset, size, NULL, NULL);
}

long font_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags;
    char* styles[] = {"UNK", "bold", "italic"};

    printf("FONT record\n");
    uint8_t tface = NEXT_BYTE();
    printf("\ttface number: %d (0x%02x)\n", tface, tface);
    uint8_t style = NEXT_BYTE();
    printf("\tstyle: %s (%d)\n", style <= 2 ? styles[style] : "UNK", style);
    printf("\tname: ");

    long name_len = 0;
    char printable = 1;
    long name_offset = offset;
    long i;
    for (i = offset; i < size && buffer[i]; i++)
    {
        name_len++;
        printable = printable && (buffer[i] >= 0x20 && buffer[i] <= 0x80);
    }

    if (buffer[i])
    {
        fprintf(stderr, "File is too small, found EOF while reading the null-terminated font name at offset 0x%lx.\n", name_offset);
        exit(2);
    }

    for (; offset < size && buffer[offset]; offset++)
        printf(printable ? "%c" : "0x%02x ", buffer[offset]);
    printf("\n");

    //Vulnerable?
    if (name_len > 0x28)
    {
        printf("\tVULNERABLE FONT RECORD FOUND!\n");
        int is_ret = 0;
        uint32_t ret_addr = *(uint32_t*)(buffer+name_offset+0x2c);
        printf("\tReturn address: 0x%x\n", ret_addr);

        for (int i = 0; i < n_rets && is_ret == 0; i++)
            is_ret = rets[i] ==  ret_addr - 0x400000;

        if (!is_ret)
        {
            printf("\tEXPLOIT NOT SUPPORTED.\n");
            fprintf(stderr, "The payload return address is not pointing to a ret instruction. Further processing of the payload is aborted.\n");
            return -1;
        }

        long host_size;
        uint8_t* host = read_file("host.exe", &host_size);
        if (! host)
            goto done;

        memcpy(host + 0x200, buffer + name_offset, 0x30);
        memcpy(host + 0x230, buffer + 4, size - 4);

        FILE* fout = fopen(output_file, "wb");
        if ( ! fout)
        {
            fprintf(stderr, "Cannot create file %s: %s (%d)\n", output_file, strerror(errno), errno);
            goto done;
        }

        if (fwrite(host, host_size, 1, fout) != 1)
        {
            fprintf(stderr, "Cannot write to file %s: %s (%d)\n", output_file, strerror(errno), errno);
            goto done;            
        }

        printf("\tHost program written to %s\n", output_file);

done:
        if (host)
            free(host);

        if (fout)
            fclose(fout);

        return -1;
    }

    return offset;
}


long size_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags;

    char* lsizes[] = {
        "szFULL", "szSUB","szSUB2","szSYM",
        "szSUBSYM","szUSER1","szUSER2"
    };

    printf("SIZE record\n");
    uint8_t second = NEXT_BYTE();
    if (second == 101)
    {
        printf("\texplicit lsize: %d\n", -NEXT_WORD());
    }
    else if (second == 100)
    {
        uint8_t lsize = NEXT_BYTE();
        uint16_t dsize = NEXT_WORD();
        printf("\tlsize: %s (%d)\n\tdsize: %d\n", lsize <= 6 ? lsizes[lsize] : "UNK", lsize, dsize);
    }
    else 
    {
        uint8_t lsize = NEXT_BYTE();
        uint16_t dsize = NEXT_BYTE() - 0x80;
        printf("\tlsize: %s (%d)\n\tdsize: %d\n", lsize <= 6 ? lsizes[lsize] : "UNK", lsize, dsize);
    }

    return offset;
}
long full_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags; (void)buffer; (void)size;
    printf("FULL record\n");
    return offset;
}

long sub_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags; (void)buffer; (void)size;
    printf("SUB record\n");
    return offset;
}

long sub2_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags; (void)buffer; (void)size;
    printf("SUB2 record\n");
    return offset;
}

long sym_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags; (void)buffer; (void)size;
    printf("SYM record\n");
    return offset;
}

long subsym_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags; (void)buffer; (void)size;
    printf("SUBSYM record\n");
    return offset;
}

long unk_record(uint8_t flags, uint8_t* buffer, long offset, long size)
{
    (void)flags; (void)buffer; (void)size; (void)offset;
    printf("UNK record\n");
    fprintf(stderr, "Found unknown record, aborting.");
    return -1;
}



long parse_mtfe(uint8_t* buffer, long offset, long size, rec_fun_t* records)
{
    //We now read the MTEF until the END record or a malformed FONT record
    while (offset >= 0)
    {
        uint8_t record = NEXT_BYTE();
        uint8_t rec_type = record & 0xf, rec_flags = record >> 4;

        offset = records[rec_type](rec_flags, buffer, offset, size);
    }

    return offset;
}



int main(int argc, char** argv)
{
    //Return value
    int retval = 0;

    //Arguments
    if (argc != 3)
        return fprintf(stderr, "Usage:\n\t%s INPUT OUTPUT\n", argv[0]), 1;
    output_file = argv[2];


    //Read the file
    long size;
    uint8_t* buffer = read_file(argv[1], &size);

    //Helpers for consuming the file
    long offset = 0;



    //Size
    long ole_size = (long)NEXT_DWORD();
    if (ole_size == size - 4)
        printf("OLE size: 0x%lx (%ld), correct\n", ole_size, ole_size);
    else
        printf("OLE size: 0x%lx (%ld), differs from file size of 0x%lx (%ld).\n\tThis won't prevent the payload from working.\n", ole_size, ole_size, size, size);

    //Header
    uint8_t header_version = NEXT_BYTE();
    printf("MTEF header:\n\tversion: %d\n", header_version);
    if (header_version != 2 && header_version != 3)
    {
        fprintf(stderr, "MTEF header version is not recognized. Only versions 2 and 3 are supported.\n");
        retval = 3;
        goto end;
    }
    uint8_t platform = NEXT_BYTE(), product = NEXT_BYTE(), major = NEXT_BYTE(), minor = NEXT_BYTE();
    printf("\tplatform: %d\n\tproduct: %d\n\tversion: %d.%d\n\tPlatform and product information are often made up.\n", platform, product, major, minor);

    
    //Parse MTFE
    parse_mtfe(buffer, offset, size, stdrec);
    

end:
    free(buffer);

    return retval;
}